import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

meta = {}
with open(os.path.join(here, 'src/masonite', '__version__.py'), 'r') as f:
    exec(f.read(), meta)

try:
    with open('README.md', 'r') as f:
        readme = f.read()
except FileNotFoundError:
    readme = ""
    
setup(
    name=meta['__title__'],
    packages=[
        'masonite',
        'masonite.auth',
        'masonite.providers',
        'masonite.listeners',
        'masonite.commands',
        'masonite.controllers',
        'masonite.drivers',
        'masonite.drivers.authentication',
        'masonite.drivers.mail',
        'masonite.drivers.broadcast',
        'masonite.drivers.cache',
        'masonite.drivers.queue',
        'masonite.drivers.session',
        'masonite.drivers.storage',
        'masonite.drivers.upload',
        'masonite.managers',
        'masonite.testsuite',
        'masonite.queues',
        'masonite.contracts',
        'masonite.contracts.managers',
        'masonite.helpers',
        'masonite.middleware',
        'masonite.testing',
    ],
    version=meta['__version__'],
    install_requires=[
        'bcrypt>=3.1,<3.2',
        'cleo>=0.6,<0.7',
        'cryptography>=2.3,<2.8',
        'hupper>=1.0,<2.0',
        'Jinja2>=2,<3',
        # 'masonite-events>=1.0,<2',
        # 'masonite-scheduler>=1.0.0,<=1.0.99',
        # 'masonite-validation>=3.0.0,<4.0.0',
        'orator>=0.9,<1',
        'passlib>=1.7,<1.8',
        'pendulum>=1.5,<1.6',
        'psutil>=5.4,<5.7',
        'python-dotenv>=0.8,<0.11',
        'requests>=2.0,<2.99',
        'tabulate>=0.8,<0.9',
        'tldextract>=2.2,<2.3',
        'whitenoise>=3.3',
    ],
    description=meta['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=meta['__author__'],
    author_email=meta['__author_email__'],
    package_dir={'': 'src'},
    url=meta['__url__'],
    keywords=['masonite', 'python web framework', 'python3'],
    licence=meta['__licence__'],
    python_requires=">=3.5",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Masonite',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    include_package_data=True,
)



"""SMTP Driver Module."""

import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from ...contracts.MailContract import MailContract
from ...drivers import BaseMailDriver


class MailSmtpDriver(BaseMailDriver, MailContract):
    """Mail smtp driver."""

    def send(self, message_contents=None):
        """Send the message through SMTP.

        Keyword Arguments:
            message {string} -- The message to be sent to SMTP. (default: {None})

        Returns:
            None
        """
        config = self.config.DRIVERS['smtp']

        message = MIMEMultipart('alternative')

        if not message_contents:
            message_contents = self.message_body

        message_contents = MIMEText(message_contents, 'html')

        message['Subject'] = self.message_subject
        message['From'] = '{0} <{1}>'.format(
            self.config.FROM['name'], self.config.FROM['address'])
        message['To'] = self.to_address
        message['Reply-To'] = self.message_reply_to
        message.attach(message_contents)

        # Send the message via our own SMTP server.
        if 'ssl' in config and config['ssl'] is True:
            self.smtp = smtplib.SMTP_SSL('{0}:{1}'.format(config['host'], config['port']))
        else:
            self.smtp = smtplib.SMTP('{0}:{1}'.format(
                config['host'], config['port']))

        self.smtp.login(config['username'], config['password'])

        if self._queue:
            from wsgi import container
            from .. import Queue
            container.make(Queue).push(
                self._send_mail,
                args=(self.config.FROM['name'], self.to_address, message.as_string())
            )
            return

        self._send_mail(self.config.FROM['name'],
                self.to_address, message.as_string())

    def _send_mail(self, *args):
        """Wrapper around sending mail so it can also be used for queues."""
        self.smtp.sendmail(*args)
        self.smtp.quit()



from orator.migrations import Migration


class CreateFailedJobsTable(Migration):

    def up(self):
        """Run the migrations."""
        with self.schema.create('failed_jobs') as table:
            table.increments('id')
            table.string('driver')
            table.string('channel')
            table.binary('payload')
            table.timestamp('failed_at')
            table.timestamps()

    def down(self):
        """Revert the migrations."""
        self.schema.drop('failed_jobs')



"""An AppProvider Service Provider."""

from ..auth import Auth
from ..autoload import Autoload
from ..commands import (AuthCommand, CommandCommand, ControllerCommand,
                               DownCommand, InfoCommand, InstallCommand,
                               JobCommand, KeyCommand, MakeMigrationCommand,
                               MiddlewareCommand, MigrateCommand,
                               MigrateRefreshCommand, MigrateResetCommand,
                               MigrateRollbackCommand, MigrateStatusCommand,
                               ModelCommand, ModelDocstringCommand,
                               ProviderCommand, PublishCommand,
                               QueueTableCommand, QueueWorkCommand,
                               RoutesCommand, SeedCommand, SeedRunCommand,
                               ServeCommand, TestCommand, TinkerCommand, UpCommand,
                               ViewCommand)
from ..exception_handler import DumpHandler, ExceptionHandler
from ..helpers import config
from ..helpers.routes import flatten_routes
from ..hook import Hook
from ..provider import ServiceProvider
from ..request import Request
from ..response import Response
from ..routes import Route

from ..managers import AuthManager
from ..drivers import AuthCookieDriver, AuthJwtDriver
from ..helpers import load


class AppProvider(ServiceProvider):

    def register(self):
        self.app.bind('HookHandler', Hook(self.app))
        self.app.bind('WebRoutes', flatten_routes(load('routes.web.routes')))
        # self.app.bind('Storage', config('storage'))
        self.app.bind('Route', Route())
        self.app.bind('Request', Request())
        self.app.simple(Response(self.app))
        self.app.bind('Container', self.app)
        self.app.bind('ExceptionHandler', ExceptionHandler(self.app))
        self.app.bind('ExceptionDumpExceptionHandler', DumpHandler)
        self.app.bind('AuthCookieDriver', AuthCookieDriver)
        self.app.bind('AuthJwtDriver', AuthJwtDriver)
        self.app.bind('AuthManager', AuthManager(self.app).driver('cookie'))
        self.app.bind('RouteMiddleware', config('middleware.route_middleware'))
        self.app.bind('HttpMiddleware', config('middleware.http_middleware'))
        self.app.bind('Auth', Auth)
        self.app.bind('staticfiles', config('storage.staticfiles', {}))

        # Insert Commands
        self._load_commands()

        self._autoload(config('application.autoload'))

    def boot(self, request: Request, route: Route):
        self.app.bind('StatusCode', None)
        route.load_environ(self.app.make('Environ'))
        request.load_environ(self.app.make('Environ')).load_app(self.app)

    def _autoload(self, directories):
        Autoload(self.app).load(directories)

    def _load_commands(self):
        self.app.bind('MasoniteAuthCommand', AuthCommand())
        self.app.bind('MasoniteCommandCommand', CommandCommand())
        self.app.bind('MasoniteControllerCommand', ControllerCommand())
        self.app.bind('MasoniteDownCommand', DownCommand())
        self.app.bind('MasoniteInfoCommand', InfoCommand())
        self.app.bind('MasoniteInstallCommand', InstallCommand())
        self.app.bind('MasoniteJobCommand', JobCommand())
        self.app.bind('MasoniteKeyCommand', KeyCommand())
        self.app.bind('MasoniteMakeMigrationCommand', MakeMigrationCommand())
        self.app.bind('MasoniteMiddlewareCommand', MiddlewareCommand())
        self.app.bind('MasoniteMigrateCommand', MigrateCommand())
        self.app.bind('MasoniteMigrateRefreshCommand', MigrateRefreshCommand())
        self.app.bind('MasoniteMigrateResetCommand', MigrateResetCommand())
        self.app.bind('MasoniteMigrateStatusCommand', MigrateStatusCommand())
        self.app.bind('MasoniteMigrateRollbackCommand',
                      MigrateRollbackCommand())
        self.app.bind('MasoniteModelCommand', ModelCommand())
        self.app.bind('MasoniteModelDocstringCommand', ModelDocstringCommand())
        self.app.bind('MasoniteProviderCommand', ProviderCommand())
        self.app.bind('MasonitePublishCommand', PublishCommand())
        self.app.bind('MasoniteQueueWorkCommand', QueueWorkCommand())
        self.app.bind('MasoniteQueueTableCommand', QueueTableCommand())
        self.app.bind('MasoniteViewCommand', ViewCommand())
        self.app.bind('MasoniteRoutesCommand', RoutesCommand())
        self.app.bind('MasoniteServeCommand', ServeCommand())
        self.app.bind('MasoniteSeedCommand', SeedCommand())
        self.app.bind('MasoniteSeedRunCommand', SeedRunCommand())
        self.app.bind('MasoniteTestCommand', TestCommand())
        self.app.bind('MasoniteTinkerCommand', TinkerCommand())
        self.app.bind('MasoniteUpCommand', UpCommand())



import os
import pathlib
import shutil

from ... import Upload
from ...contracts import StorageContract
from ...drivers import BaseDriver


class StorageDiskDriver(BaseDriver, StorageContract):

    def put(self, location, contents):
        with open(location, 'w+') as file:
            file.write(contents)

    def append(self, location, contents):
        with open(location, 'a+') as file:
            file.write(contents)

    def get(self, location):
        with open(location) as f:
            return f.read()

    def delete(self, location):
        try:
            os.remove(location)
            return True
        except FileNotFoundError:
            return False

    def exists(self, location):
        return pathlib.Path(location).exists()

    def size(self, location):
        try:
            return os.path.getsize(location)
        except FileNotFoundError:
            return 0

    def extension(self, location):
        return pathlib.Path(location).suffix.replace('.', '')

    def url(self, location):
        pass

    def name(self, location):
        return pathlib.Path(location).name

    def upload(self, *args, **kwargs):
        from wsgi import container
        return container.make(Upload).driver('disk').store(*args, **kwargs)

    def all(self):
        pass

    def make_directory(self, location):
        location = os.path.join(os.getcwd(), location)
        if os.path.isdir(location):
            return True

        os.mkdir(location)
        return True

    def delete_directory(self, directory, force=False):
        if force:
            shutil.rmtree(directory)
            return True
        try:
            pathlib.Path(directory).rmdir()
            return True
        except FileNotFoundError:
            return True

    def move(self, old, new):
        return shutil.move(old, new)




import time
import os

from hupper.logger import DefaultLogger, LogLevel
from hupper.reloader import Reloader, find_default_monitor_factory
from cleo import Command
from ..helpers import has_unmigrated_migrations
from ..exceptions import DriverLibraryNotFound


class ServeCommand(Command):
    """
    Run the Masonite server.

    serve
        {--p|port=8000 : Specify which port to run the server}
        {--b|host=127.0.0.1 : Specify which ip address to run the server}
        {--r|reload : Make the server automatically reload on file changes}
        {--d|dont-reload : Make the server NOT automatically reload on file changes}
        {--i|reload-interval=1 : Number of seconds to wait to reload after changed are detected}
        {--l|live-reload : Make the server automatically refresh your web browser}
    """

    def handle(self):
        if has_unmigrated_migrations():
            self.comment("\nYou have unmigrated migrations. Run 'craft migrate' to migrate them\n")

        if self.option('live-reload'):
            try:
                from livereload import Server
            except ImportError:
                raise DriverLibraryNotFound("Could not find the livereload library. Install it by running 'pip install livereload==2.5.1'")

            from wsgi import container
            from config import application
            import glob

            server = Server(container.make('WSGI'))
            for filepath in glob.glob('resources/templates/**/*/'):
                server.watch(filepath)

            self.line('')
            self.info('Live reload server is starting...')
            self.info(
                'This will only work for templates. Changes to Python files may require a browser refresh.')
            self.line('')
            application = server.serve(port=self.option('port'), restart_delay=self.option(
                'reload-interval'), liveport=5500, root=application.BASE_DIRECTORY, debug=True)
            return

        if not self.option('dont-reload'):
            logger = DefaultLogger(LogLevel.INFO)

            # worker args are pickled and then passed to the new process
            worker_args = [
                self.option("host"), self.option("port"), "wsgi:application",
            ]

            reloader = Reloader(
                "masonite.commands._devserver.run",
                find_default_monitor_factory(logger),
                logger,
                worker_args=worker_args,
            )

            self._run_reloader(reloader, extra_files=[".env", "storage/"])

        else:
            from wsgi import application
            from ._devserver import run
            run(self.option("host"), self.option("port"), application)

    def _run_reloader(self, reloader, extra_files=[]):
        reloader._capture_signals()
        reloader._start_monitor()
        for blob in extra_files:
            reloader.monitor.add_path(os.path.join(os.getcwd(), blob))
        try:
            while True:
                if not reloader._run_worker():
                    reloader._wait_for_changes()
                time.sleep(float(self.option('reload-interval')))
        except KeyboardInterrupt:
            pass
        finally:
            reloader._stop_monitor()
            reloader._restore_signals()



"""Async Driver Method."""

import inspect
import pickle
import time

import pendulum
from ...contracts import QueueContract
from ...drivers import BaseQueueDriver
from ...helpers import HasColoredCommands, parse_human_time
from ...queues import Queueable


class QueueDatabaseDriver(BaseQueueDriver, HasColoredCommands, QueueContract):
    """Queue Aysnc Driver."""

    def __init__(self):
        """Queue Async Driver.

        Arguments:
            Container {masonite.app.App} -- The application container.
        """
        pass

    def connect(self):
        return self

    def push(self, *objects, args=(), kwargs={}, **options):
        """Push objects onto the async stack.

        Arguments:
            objects {*args of objects} - This can be several objects as parameters into this method.
            options {**kwargs of options} - Additional options for async driver
        """

        from config.database import DB as schema

        callback = options.get('callback', 'handle')
        wait = options.get('wait', None)
        connection = options.get('connection', None)

        if connection:
            schema = schema.connection(connection)

        if wait:
            wait = parse_human_time(wait).to_datetime_string()

        for job in objects:
            if schema.get_schema_builder().has_table('queue_jobs'):
                payload = pickle.dumps({'obj': job, 'args': args, 'callback': callback})
                schema.table('queue_jobs').insert({
                    'name': str(job),
                    'serialized': payload,
                    'created_at': pendulum.now().to_datetime_string(),
                    'attempts': 0,
                    'ran_at': None,
                    'wait_until': wait,
                })

    def consume(self, channel, fair=False, **options):
        from config.database import DB as schema, DATABASES
        from wsgi import container

        if not channel or channel == 'default':
            channel = DATABASES['default']

        self.info('[*] Waiting to process jobs from the "queue_jobs" table on the "{}" connection. To exit press CTRL + C'.format(channel))
        schema = schema.connection(channel)
        while True:
            jobs = schema.table('queue_jobs').where('ran_at', None).get()
            if not jobs.count():
                time.sleep(5)

            for job in jobs:
                unserialized = pickle.loads(job.serialized)
                obj = unserialized['obj']
                args = unserialized['args']
                callback = unserialized['callback']
                ran = job.attempts

                wait_time = job['wait_until']

                if isinstance(wait_time, str):
                    wait_time = pendulum.parse(job['wait_until'])
                else:
                    wait_time = pendulum.instance(job['wait_until'])

                # print(job['wait_until'], wait_time.is_future())
                if job['wait_until'] and wait_time.is_future():
                    continue
                try:
                    try:
                        if inspect.isclass(obj):
                            obj = container.resolve(obj)

                        getattr(obj, callback)(*args)

                    except AttributeError:
                        obj(*args)

                    try:
                        # attempts = 1
                        schema.table('queue_jobs').where('id', job['id']).update({
                            'ran_at': pendulum.now().to_datetime_string(),
                            'attempts': job['attempts'] + 1,
                        })
                        self.success('[\u2713] Job Successfully Processed')
                    except UnicodeEncodeError:
                        self.success('[Y] Job Successfully Processed')
                except Exception as e:
                    self.danger('Job Failed: {}'.format(str(e)))

                    if not obj.run_again_on_fail:
                        # ch.basic_ack(delivery_tag=method.delivery_tag)
                        schema.table('queue_jobs').where('id', job['id']).update({
                            'ran_at': pendulum.now().to_datetime_string(),
                            'failed': 1,
                            'attempts': job['attempts'] + 1,
                        })

                    if ran < obj.run_times and isinstance(obj, Queueable):
                        time.sleep(1)
                        schema.table('queue_jobs').where('id', job['id']).update({
                            'attempts': job['attempts'] + 1,
                        })
                        continue
                    else:
                        schema.table('queue_jobs').where('id', job['id']).update({
                            'attempts': job['attempts'] + 1,
                            'ran_at': pendulum.now().to_datetime_string(),
                            'failed': 1,
                        })

                        if hasattr(obj, 'failed'):
                            getattr(obj, 'failed')(unserialized, str(e))

                        self.add_to_failed_queue_table(unserialized, driver='database')

            time.sleep(5)



"""A SassProvider Service Provider."""

from ..provider import ServiceProvider
from ..storage import Storage


class SassProvider(ServiceProvider):

    wsgi = False

    def register(self):
        """Compile Sass.

        Compile Sass if the libsass module is installed. Once installed, all
        Sass files are compiled when the server is ran. This will only run
        once when the server is first started.
        """
        Storage().compile_sass()

    def boot(self):
        pass
